/*
 * Copyright (c) 2006-2009 Erez Zadok
 * Copyright (c) 2006      Charles P. Wright
 * Copyright (c) 2006-2007 Josef 'Jeff' Sipek
 * Copyright (c) 2006      Junjiro Okajima
 * Copyright (c) 2006      David P. Quigley
 * Copyright (c) 2006-2009 Stony Brook University
 * Copyright (c) 2006-2009 The Research Foundation of SUNY
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "union.h"

/*
 * Super-user IO work Queue - sometimes we need to perform actions which
 * would fail due to the unix permissions on the parent directory (e.g.,
 * rmdir a directory which appears empty, but in reality contains
 * whiteouts).
 */

struct workqueue_struct *sioq;

int __init init_sioq(void)
{
	int err;

	sioq = create_workqueue("unionfs_siod");
	if (!IS_ERR(sioq))
		return 0;

	err = PTR_ERR(sioq);
	printk(KERN_ERR "create_workqueue failed %d\n", err);
	sioq = NULL;
	return err;
}

void stop_sioq(void)
{
	if (sioq)
		destroy_workqueue(sioq);
}

void run_sioq(void (*func)(void *arg), struct sioq_args *args)
{
	DECLARE_WORK(wk, func, &args->comp);

	init_completion(&args->comp);
	while (!queue_work(sioq, &wk)) {
		/* TODO: do accounting if needed */
		schedule();
	}
	wait_for_completion(&args->comp);
}

void __unionfs_create(void *data)
{
	struct sioq_args *args = data;

	args->err = vfs_create(args->create.parent, args->create.dentry,
				args->create.mode, args->create.nd);
	complete(&args->comp);
}

void __unionfs_mkdir(void *data)
{
	struct sioq_args *args = data;

	args->err = vfs_mkdir(args->mkdir.parent, args->mkdir.dentry,
				args->mkdir.mode);
	complete(&args->comp);
}

void __unionfs_mknod(void *data)
{
	struct sioq_args *args = data;

	args->err = vfs_mknod(args->mknod.parent, args->mknod.dentry,
				args->mknod.mode, args->mknod.dev);
	complete(&args->comp);
}

void __unionfs_symlink(void *data)
{
	struct sioq_args *args = data;

	args->err = vfs_symlink(args->symlink.parent, args->symlink.dentry,
				args->symlink.symbuf, args->symlink.mode);
	complete(&args->comp);
}

void __unionfs_unlink(void *data)
{
	struct sioq_args *args = data;

	args->err = vfs_unlink(args->unlink.parent, args->unlink.dentry);
	complete(&args->comp);
}
